/*****************************************************************************

    Bible Reader for Palm
    Copyright (C) 2001  Ho Chuen Poon
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    
    You may contact Ho Chuen Poon at poetry@hongkong.com

*****************************************************************************/

#include <PalmOS.h>
#include "PalmBible.h"

// As PalmOS is single threaded, make it global and reserve names
// for cleaner coding. No parameters passing between internal functions.

static ReaderPtr rdrReader;

// Private Functions

static UInt32 getValue32Start(UInt32* valuePtr, Int16 index)
{
	UInt32 value;

	if (index > 0){
		value = valuePtr[index - 1];
	} else {
		value = 0;
	}
	
	return value;
}

static UInt16 getValue16Start(UInt16* valuePtr, Int16 index)
{
	if (index > 0){
		return valuePtr[index - 1];
	} else {
		return 0;
	}
}

static UInt16 getValue16Length(UInt16* valuePtr, Int16 index)
{
	Int16 diff;
	
	diff = getValue16Start(valuePtr, index + 1) - getValue16Start(valuePtr, index);
	if (diff < 0){
		diff = 0;
	}
	return diff; 
}

static UInt16 getTotalChapter()
{
	return rdrReader->totalChapters;
}

static UInt16 getTotalVerse()
{
	return getValue16Length(rdrReader->totalVersesAcc, rdrReader->chapter - 1);
}

static UInt16 getVerseStartLength()
{
	UInt16 verseAcc;

	verseAcc = getValue16Start(rdrReader->totalVersesAcc, 
			rdrReader->chapter - 1) + rdrReader->verse;
	rdrReader->verseStart = getValue32Start(rdrReader->totalChapterCharsAcc, 
			rdrReader->chapter);
	
	if (rdrReader->verse > 1){
		rdrReader->verseStart += getValue16Start(rdrReader->totalVerseCharsAcc, 
				verseAcc - 1);
		rdrReader->verseLength = getValue16Length(rdrReader->totalVerseCharsAcc, 
			verseAcc - 1);
	} else {
		rdrReader->verseLength = getValue16Start(rdrReader->totalVerseCharsAcc, 
			verseAcc);
	}
}

static openIndexRecord()
{
	UInt16* tempPtr;

	if (rdrReader->version != 0 && rdrReader->indexRecord != 0){
		tempPtr = DmxLockVersionRecord(rdrReader->version, rdrReader->indexRecord);

		rdrReader->totalChapters = tempPtr[0];
		rdrReader->totalVersesAcc = &tempPtr[1];
		rdrReader->totalChapterCharsAcc = (UInt32*) &tempPtr[rdrReader->totalChapters + 1];
		rdrReader->totalVerseCharsAcc = &tempPtr[rdrReader->totalChapters * 3 + 1];
	}
}

static closeIndexRecord()
{
	if (rdrReader->version != 0 && rdrReader->indexRecord != 0){
		DmxUnlockVersionRecord(rdrReader->version, rdrReader->indexRecord);
	}
}

static void closeAllReader()
{
	if (rdrReader->version != 0 && rdrReader->indexRecord != 0) {
		closeIndexRecord();
		DecCloseDecoder(&rdrReader->decoder);
		FrdCloseReader(&rdrReader->recordReader);
	}
}

// public functions

void RdrOpenReader(ReaderPtr reader)
{
	reader->version = 0;
}

void RdrCloseReader(ReaderPtr reader)
{
	rdrReader = reader;
	closeAllReader();
}

void RdrBookmark(ReaderPtr reader)
{
	reader->bookmarkChapter = reader->chapter; 
	reader->bookmarkVerse = reader->verse; 
}

void RdrReturn(ReaderPtr reader)
{
	reader->chapter = reader->bookmarkChapter;
	reader->verse = reader->bookmarkVerse;
}

Boolean RdrSeekVerse(ReaderPtr reader, UInt16 version, UInt16 book, UInt16 chapter, UInt16 verse)
{
	DmOpenRef versionDB;
	UInt16 totalRecord;
	
	rdrReader = reader;

	if (version != rdrReader->version || book != rdrReader->book){
		closeAllReader();

		rdrReader->indexRecord = InfGetBookRecordIndex(version, book);
		
		if (rdrReader->indexRecord != 0){
			rdrReader->totalRecord = InfGetBookTotalRecord(version, book);
		
			DecOpenDecoder(&rdrReader->decoder, version);
			FrdOpenReader(&rdrReader->recordReader, version, rdrReader->indexRecord + 1, rdrReader->indexRecord + rdrReader->totalRecord);
			rdrReader->version = version;
			rdrReader->book = book;

			openIndexRecord();
		} else {
			rdrReader->chapter = 0;
			rdrReader->verse = 0;

			return false;
		}
	}

	rdrReader->chapter = chapter;
	rdrReader->verse = verse;
	
	return true;
}

VersePtr RdrGetCurrentVerse(ReaderPtr reader)
{
	VersePtr versePtr;

	rdrReader = reader;
	
	if (rdrReader->chapter != 0 && rdrReader->chapter < getTotalChapter() + 1){
		versePtr = VersePtrNew();
		versePtr->chapter = rdrReader->chapter;
		versePtr->verse = rdrReader->verse;
		versePtr->next = NULL;
		versePtr->prev = NULL;

		getVerseStartLength();
		if (rdrReader->verseLength > 0){
			DecDecodeVerse(&rdrReader->decoder, versePtr, &rdrReader->recordReader, rdrReader->verseStart, rdrReader->verseLength);
		} else {
			versePtr->lines = 0;
		}

		return versePtr;
	} else {
		return NULL;
	}
}

Boolean RdrGetCurrentVerseAsString(ReaderPtr reader, char* versePtr)
{
	rdrReader = reader;
	
	if (rdrReader->chapter != 0 && rdrReader->chapter < getTotalChapter() + 1){
		getVerseStartLength();
		if (rdrReader->verseLength > 0){
			DecDecodeVerseAsString(&rdrReader->decoder, versePtr, &rdrReader->recordReader, rdrReader->verseStart, rdrReader->verseLength);
		} else {
			versePtr[0] = 0;
		}
	} else {
		return false;
	}
	
	return true;
}

void RdrMovePrevVerse(ReaderPtr reader)
{
	rdrReader = reader;

	rdrReader->verse --;
	if (rdrReader->verse < 1){
		if (rdrReader->chapter > 0){
			rdrReader->chapter --;
			rdrReader->verse = getTotalVerse();
		}
	}
}

void RdrMoveNextVerse(ReaderPtr reader)
{
	rdrReader = reader;

	rdrReader->verse ++;
	if (rdrReader->verse > getTotalVerse()) {
		if (rdrReader->chapter < getTotalChapter() + 1) {
			rdrReader->chapter ++;
			rdrReader->verse = 1;
		}
	}
}

